@@ -166,7 +166,8 @@
 				if (tree.data('ztfy-tree-details') == 'off') {
 					return;
 				}
-				var term = $(source).text().replace(/ /g, '%20');
+				var term = $(source).text().replace(/ /g, '%20')
+										   .replace(/&/g, '%26');
 				if ($.browser.msie) {
 					term = $.UTF8.encode(term);
 				}
@@ -178,7 +179,8 @@
 				if (source instanceof $.Event) {
 					var source  = source.srcElement || source.target;
 				}
-				var term = $(source).text().replace(/ /g, '%20');
+				var term = $(source).text().replace(/ /g, '%20')
+										   .replace(/&/g, '%26');
 				if (term) {
 					if ($.browser.msie) {
 						term = $.UTF8.encode(term);
