@@ -17,6 +17,7 @@
 # import standard packages
 from persistent import Persistent
 import BTrees.Length
+import re
 
 # import Zope3 interfaces
 from hurry.query.interfaces import IQuery
@@ -28,6 +29,7 @@ from zope.container.interfaces import IObjectAddedEvent, IObjectRemovedEvent
 from zope.dublincore.interfaces import IZopeDublinCore
 from zope.location.interfaces import ISublocations
 from zopyx.txng3.core.interfaces.ting import ITingIndex
+from zopyx.txng3.core.parsers.english import QueryParserError
 
 # import local interfaces
 from ztfy.thesaurus.interfaces.loader import IThesaurusLoader
@@ -38,7 +40,7 @@ from ztfy.thesaurus.interfaces.thesaurus import IThesaurus, \
 from ztfy.thesaurus.interfaces.tree import ITree
 
 # import Zope3 packages
-from hurry.query import And
+from hurry.query import And, Or
 from hurry.query.value import Eq
 from zc.catalog.catalogindex import ValueIndex
 from zope.catalog.catalog import Catalog
@@ -59,6 +61,9 @@ from ztfy.utils.traversing import getParent
 from ztfy.utils.unicode import translateString
 
 
+CUSTOM_SEARCH = re.compile(r'\*|\"|\sAND\s|\sOR\s|\sNOT\s|\(|\)', re.IGNORECASE)
+
+
 class ThesaurusIndexSearch(BaseTextSearch):
     """Custom text search index using catalog instance instead of name"""
 
@@ -206,6 +211,7 @@ class Thesaurus(Persistent, Contained):
         index = TextIndexNG('label', IThesaurusTerm,
                             field_callable=False,
                             languages=self.language,
+                            splitter='txng.splitters.default',
                             storage='txng.storages.term_frequencies',
                             dedicated_storage=False,
                             use_stopwords=True,
@@ -213,6 +219,19 @@ class Thesaurus(Persistent, Contained):
                             ranking=True)
         locate(index, catalog, 'label')
         catalog['label'] = index
+        # init stemmed search catalog
+        index = TextIndexNG('label', IThesaurusTerm,
+                            field_callable=False,
+                            languages=self.language,
+                            splitter='txng.splitters.default',
+                            storage='txng.storages.term_frequencies',
+                            dedicated_storage=False,
+                            use_stopwords=True,
+                            use_normalizer=True,
+                            use_stemmer=True,
+                            ranking=True)
+        locate(index, catalog, 'stemm_label')
+        catalog['stemm_label'] = index
         # init basic label catalog
         index = ValueIndex('base_label', IThesaurusTerm, field_callable=False)
         locate(index, catalog, 'base_label')
@@ -222,7 +241,11 @@ class Thesaurus(Persistent, Contained):
         for index, term in enumerate(self.terms.itervalues()):
             indexObject(term, catalog)
             if not index % 100:
-                ITransactionManager(catalog).savepoint()
+                try:
+                    ITransactionManager(catalog).savepoint()
+                except TypeError:
+                    # This method can fail if thesaurus is not stored yet...
+                    pass
 
     def delete(self):
         manager = getSiteManager(self)
@@ -265,7 +288,11 @@ class Thesaurus(Persistent, Contained):
                 elif not IThesaurusLoaderTerm.providedBy(term):
                     terms[key] = term
                 if not index % 100:
-                    ITransactionManager(self).savepoint()
+                    try:
+                        ITransactionManager(self).savepoint()
+                    except TypeError:
+                        # This method can fail if thesaurus is not stored yet...
+                        pass
         self.resetTermsParent()
         self.resetTopTerms()
 
@@ -281,14 +308,18 @@ class Thesaurus(Persistent, Contained):
                 parent = parent.generic
             term.parent = parent
             if not index % 100:
-                ITransactionManager(self).savepoint()
+                try:
+                    ITransactionManager(self).savepoint()
+                except TypeError:
+                    # This method can fail if thesaurus is not stored yet...
+                    pass
 
     def resetTopTerms(self):
         self.top_terms = [term for term in self.terms.itervalues()
                           if (not term.generic) and (not term.usage)]
 
-    def findTerms(self, query=None, extract=None, autoexpand='on_miss', glob='end', limit=None,
-                  exact=False, exact_only=False):
+    def findTerms(self, query=None, extract=None, autoexpand='always', glob='end', limit=None,
+                  exact=False, exact_only=False, stemmed=False):
         assert exact or (not exact_only)
         query_util = getUtility(IQuery)
         terms = []
@@ -297,22 +328,35 @@ class Thesaurus(Persistent, Contained):
                                           translateString(query, escapeSlashes=True, forceLower=True, spaces=' '))
             terms = list(query_util.searchResults(search))
         if not exact_only:
+            searches = []
+            # check stemmed index
+            if stemmed and not re.search(r'\*', query):
+                index = (self.catalog, 'stemm_label')
+                searches.append(ThesaurusIndexSearch(index, {'query': u' '.join(m for m in query.split() if len(m) > 2),
+                                                             'autoexpand': 'off',
+                                                             'ranking': False}))
+            # check basic index
             start = ''
             end = ''
-            if '*' not in query:
+            if CUSTOM_SEARCH.search(query):
+                query_text = query
+                autoexpand = 'off'
+            else:
                 if glob in ('start', 'both'):
                     start = u'*'
                 if glob in ('end', 'both'):
                     end = u'*'
-            search = ThesaurusIndexSearch((self.catalog, 'label'), {'query': u' '.join((u'%s%s%s' % (start, m, end)
-                                                                                        for m in query.split()
-                                                                                              if len(m) > 2)),
-                                                                    'autoexpand': autoexpand,
-                                                                    'ranking': False})
-            terms += sorted(query_util.searchResults(search, limit=limit), key=lambda x: x.label)
+                query_text = u' '.join((u'%s%s%s' % (start, m, end) for m in query.split() if len(m) > 2))
+            index = (self.catalog, 'label')
+            searches.append(ThesaurusIndexSearch(index, {'query': query_text,
+                                                         'autoexpand': autoexpand,
+                                                         'ranking': False}))
+            try:
+                terms += sorted(query_util.searchResults(Or(*searches), limit=limit), key=lambda x: x.label)
+            except QueryParserError:
+                pass
         if extract:
-            terms = [term for term in terms
-                                   if extract in term.extracts]
+            terms = [term for term in terms if extract in term.extracts]
         return terms
 
 
